@@ -1,11 +1,10 @@
 #![doc(
-    html_root_url = "https://docs.rs/arc-swap/0.4.4/arc-swap/",
+    html_root_url = "https://docs.rs/arc-swap/0.4.7/arc-swap/",
     test(attr(deny(warnings)))
 )]
 #![deny(missing_docs, warnings)]
 // We aim at older rust too, one without dyn
 #![allow(unknown_lints, bare_trait_objects, renamed_and_removed_lints)]
-#![cfg_attr(feature = "unstable-weak", feature(weak_into_raw))]
 
 //! Making [`Arc`][Arc] itself atomic
 //!
@@ -221,11 +220,9 @@
 //!
 //! # Features
 //!
-//! The `unstable-weak` feature adds the ability to use arc-swap with the [Weak] pointer too,
-//! through the [ArcSwapWeak] type. This requires the nightly Rust compiler. Also, the interface
-//! and support **is not** part of API stability guarantees and may be arbitrarily changed or
-//! removed in future releases (it is mostly waiting for the `weak_into_raw` nightly feature to
-//! stabilize before stabilizing it in this crate).
+//! The `weak` feature adds the ability to use arc-swap with the [Weak] pointer too,
+//! through the [ArcSwapWeak] type. The needed std support is stabilized in rust version 1.45 (as
+//! of now in beta).
 //!
 //! # Internal details
 //!
@@ -237,6 +234,25 @@
 //! * [Making `Arc` more atomic](https://vorner.github.io/2018/06/24/arc-more-atomic.html)
 //! * [More tricks up in the ArcSwap's sleeve](https://vorner.github.io/2019/04/06/tricks-in-arc-swap.html)
 //!
+//! # Limitations
+//!
+//! This currently works only for `Sized` types. Unsized types have „fat pointers“, which are twice
+//! as large as the normal ones. The [`AtomicPtr`] doesn't support them. One could use something
+//! like `AtomicU128` for them. The catch is this doesn't exist and the difference would make it
+//! really hard to implement the debt storage/stripped down hazard pointers.
+//!
+//! A workaround is to use double indirection:
+//!
+//! ```rust
+//! # use arc_swap::ArcSwap;
+//! // This doesn't work:
+//! // let data: ArcSwap<[u8]> = ArcSwap::new(Arc::from([1, 2, 3]));
+//!
+//! // But this does:
+//! let data: ArcSwap<Box<[u8]>> = ArcSwap::from_pointee(Box::new([1, 2, 3]));
+//! # drop(data);
+//! ```
+//!
 //! [Arc]: https://doc.rust-lang.org/std/sync/struct.Arc.html
 //! [Weak]: https://doc.rust-lang.org/std/sync/struct.Arc.html
 //! [RwLock]: https://doc.rust-lang.org/std/sync/struct.RwLock.html
@@ -257,7 +273,7 @@ mod compile_fail_tests;
 mod debt;
 pub mod gen_lock;
 mod ref_cnt;
-#[cfg(feature = "unstable-weak")]
+#[cfg(feature = "weak")]
 mod weak;
 
 use std::fmt::{Debug, Display, Formatter, Result as FmtResult};
@@ -480,12 +496,23 @@ pub struct Guard<'l, T: RefCnt> {
 }
 
 impl<'a, T: RefCnt> Guard<'a, T> {
-    /// Turns the given guard into an unprotected, independent one.
+    fn new(ptr: *const T::Base, protection: Protection<'a>) -> Guard<'a, T> {
+        Guard {
+            inner: ManuallyDrop::new(unsafe { T::from_ptr(ptr) }),
+            protection,
+        }
+    }
+
+    /// Converts it into the held value.
     ///
-    /// It'll not hold any locks, will have no debts and will contain full-featured value inside
-    /// that even can outlive the ArcSwap it originated from.
+    /// This, on occasion, may be a tiny bit faster than cloning the Arc or whatever is being held
+    /// inside.
+    // Associated function on purpose, because of deref
+    #[cfg_attr(feature = "cargo-clippy", allow(wrong_self_convention))]
     #[inline]
-    fn unprotected(mut lease: Self) -> Guard<'static, T> {
+    pub fn into_inner(mut lease: Self) -> T {
+        // Drop any debt and release any lock held by the given guard and return a
+        // full-featured value that even can outlive the ArcSwap it originated from.
         match mem::replace(&mut lease.protection, Protection::Unprotected) {
             // Not protected, nothing to unprotect.
             Protection::Unprotected => (),
@@ -505,29 +532,35 @@ impl<'a, T: RefCnt> Guard<'a, T> {
                 lock.unlock();
             }
         }
+
         // The ptr::read & forget is something like a cheating move. We can't move it out, because
         // we have a destructor and Rust doesn't allow us to do that.
-        let inner = ManuallyDrop::new(unsafe { ptr::read(lease.inner.deref()) });
+        let inner = unsafe { ptr::read(lease.inner.deref()) };
         mem::forget(lease);
-        Guard {
-            inner,
-            protection: Protection::Unprotected,
-        }
+        inner
     }
 
-    /// Converts it into the held value.
+    /// Create a guard for a given value `inner`.
     ///
-    /// This, on occasion, may be a tiny bit faster than cloning the Arc or whatever is being held
-    /// inside.
-    // Associated function on purpose, because of deref
-    #[cfg_attr(feature = "cargo-clippy", allow(wrong_self_convention))]
-    pub fn into_inner(lease: Self) -> T {
-        let unprotected = Guard::unprotected(lease);
-        // The ptr::read & forget is something like a cheating move. We can't move it out, because
-        // we have a destructor and Rust doesn't allow us to do that.
-        let res = unsafe { ptr::read(unprotected.inner.deref()) };
-        mem::forget(unprotected);
-        res
+    /// This can be useful on occasion to pass a specific object to code that expects or
+    /// wants to store a Guard.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// # use arc_swap::{ArcSwap, Guard};
+    /// # use std::sync::Arc;
+    /// # let p = ArcSwap::from_pointee(42);
+    /// // Create two guards pointing to the same object
+    /// let g1 = p.load();
+    /// let g2 = Guard::from_inner(Arc::clone(&*g1));
+    /// # drop(g2);
+    /// ```
+    pub fn from_inner(inner: T) -> Self {
+        Guard {
+            inner: ManuallyDrop::new(inner),
+            protection: Protection::Unprotected,
+        }
     }
 }
 
@@ -749,12 +782,8 @@ impl<T: RefCnt, S: LockStorage> ArcSwapAny<T, S> {
     fn lock_internal(&self, signal_safe: SignalSafety) -> Guard<'_, T> {
         let gen = GenLock::new(signal_safe, &self.lock_storage);
         let ptr = self.ptr.load(Ordering::Acquire);
-        let inner = ManuallyDrop::new(unsafe { T::from_ptr(ptr) });
 
-        Guard {
-            inner,
-            protection: Protection::Lock(gen),
-        }
+        Guard::new(ptr, Protection::Lock(gen))
     }
 
     /// An async-signal-safe version of [`load`](#method.load)
@@ -787,13 +816,9 @@ impl<T: RefCnt, S: LockStorage> ArcSwapAny<T, S> {
         let debt = Debt::new(ptr as usize)?;
 
         let confirm = self.ptr.load(Ordering::Acquire);
-        let inner = ManuallyDrop::new(unsafe { T::from_ptr(ptr) });
         if ptr == confirm {
             // Successfully got a debt
-            Some(Guard {
-                inner,
-                protection: Protection::Debt(debt),
-            })
+            Some(Guard::new(ptr, Protection::Debt(debt)))
         } else if debt.pay::<T>(ptr) {
             // It changed in the meantime, we return the debt (that is on the outdated pointer,
             // possibly destroyed) and fail.
@@ -801,10 +826,7 @@ impl<T: RefCnt, S: LockStorage> ArcSwapAny<T, S> {
         } else {
             // It changed in the meantime, but the debt for the previous pointer was already paid
             // for by someone else, so we are fine using it.
-            Some(Guard {
-                inner,
-                protection: Protection::Unprotected,
-            })
+            Some(Guard::new(ptr, Protection::Unprotected))
         }
     }
 
@@ -855,8 +877,12 @@ impl<T: RefCnt, S: LockStorage> ArcSwapAny<T, S> {
     /// ```
     #[inline]
     pub fn load(&self) -> Guard<'static, T> {
-        self.load_fallible()
-            .unwrap_or_else(|| Guard::unprotected(self.lock_internal(SignalSafety::Unsafe)))
+        self.load_fallible().unwrap_or_else(|| {
+            let locked = self.lock_internal(SignalSafety::Unsafe);
+            // Extracting the object into a full-featured value has the
+            // side effect of dropping the lock.
+            Guard::from_inner(Guard::into_inner(locked))
+        })
     }
 
     /// Replaces the value inside this instance.
@@ -961,18 +987,15 @@ impl<T: RefCnt, S: LockStorage> ArcSwapAny<T, S> {
             unsafe { T::dec(new) };
         }
 
-        let inner = ManuallyDrop::new(unsafe { T::from_ptr(previous_ptr) });
-        Guard {
-            inner,
-            protection: debt.into(),
-        }
+        Guard::new(previous_ptr, debt.into())
     }
 
     /// Wait until all readers go away.
     fn wait_for_readers(&self, old: *const T::Base) {
         let mut seen_group = [false; GEN_CNT];
         let mut iter = 0usize;
-        while !seen_group.iter().all(|seen| *seen) {
+
+        loop {
             // Note that we don't need the snapshot to be consistent. We just need to see both
             // halves being zero, not necessarily at the same time.
             let gen = self.lock_storage.gen_idx().load(Ordering::Relaxed);
@@ -996,11 +1019,18 @@ impl<T: RefCnt, S: LockStorage> ArcSwapAny<T, S> {
             for i in 0..GEN_CNT {
                 seen_group[i] = seen_group[i] || (groups[i] == 0);
             }
+
+            if seen_group.iter().all(|seen| *seen) {
+                break;
+            }
+
             iter = iter.wrapping_add(1);
-            if iter % YIELD_EVERY == 0 {
-                thread::yield_now();
-            } else {
-                atomic::spin_loop_hint();
+            if cfg!(not(miri)) {
+                if iter % YIELD_EVERY == 0 {
+                    thread::yield_now();
+                } else {
+                    atomic::spin_loop_hint();
+                }
             }
         }
         Debt::pay_all::<T>(old);
@@ -1324,7 +1354,7 @@ pub type IndependentArcSwap<T> = ArcSwapAny<Arc<T>, PrivateUnsharded>;
 /// [`ArcSwapAny`](struct.ArcSwapAny.html).
 ///
 /// [Weak]: std::sync::Weak
-#[cfg(feature = "unstable-weak")]
+#[cfg(feature = "weak")]
 pub type ArcSwapWeak<T> = ArcSwapAny<std::sync::Weak<T>>;
 
 #[cfg(test)]
